/*
 * linux/fs/bfs/dir.c
 *
 * Copyright (C) 1999 Makoto Kato (m_kato@ga2.so-net.ne.jp)
 */

#include <asm/uaccess.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/bfs_fs.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/mm.h>


static ssize_t bfs_dir_read (struct file * filp, char * buf, size_t count,
	loff_t *ppos)
{
	return -EISDIR;
}


static int bfs_readdir(struct file *, void *, filldir_t);

static struct file_operations bfs_dir_operations = {
	NULL,			/* lseek - default */
	bfs_dir_read,		/* read */
	NULL,			/* write - bad */
	bfs_readdir,		/* readdir */
	NULL,			/* poll - default */
	NULL,			/* ioctl */
	NULL,			/* mmap */
	NULL,			/* no special open code */
	NULL,			/* flush */
	NULL,			/* no special release code */
	file_fsync,		/* fsync */
	NULL,			/* fasync */
	NULL,			/* check_media_change */
	NULL			/* revalidate */
};

/*
 * directories can handle most operations...
 */
struct inode_operations bfs_dir_inode_operations = {
	&bfs_dir_operations,	/* default directory file-ops */
	NULL,			/* create */
	bfs_lookup,		/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,			/* readpage */
	NULL,			/* writepage */
	NULL,			/* bmap */
	NULL,			/* truncate */
	NULL,			/* permission */
	NULL			/* smap */
};


/*
 * A directry structure of BFS become index structure.
 *
 * sturct bfs_index_node
 * {
 *   off_t  left_link;
 *   off_t  right_link;
 *   off_t  overflow_link;
 *   short  all_node_count;
 *   short  all_key_length
 *   byte   key_data[all_key_length]
 *   short  key_index_count[all_node_count-1]
 *   off_t  values[all_node_count]
 * }
 * 
 * directry have above key.
 * [.] [..] [file and directry] ...
 */

static int bfs_readdir(struct file * filp, void * dirent, filldir_t filldir)
{
	struct inode *       inode = filp->f_dentry->d_inode;
	struct super_block * sb = inode->i_sb;
	bfs_inode_addr       iaddr;
	char *               tmpname;
	int                  error;
	int                  count = 0;
	int                  len;
	bfs_data_stream *    ds;
	int                  pos = 0;
	char                 name[BFS_NAME_LEN + 1];

	BFS_OUTPUT (("---> Enter bfs_readdir "));
	BFS_OUTPUT (("inode %ld\n filp->f_pos %d\n",
		inode->i_ino, filp->f_pos));

	if (!inode || !S_ISDIR(inode->i_mode) || S_ISLNK(inode->i_mode))
		return -EBADF;

	tmpname = (char *) __getname();
	if (!tmpname)
		return -ENOMEM;

	ds = &inode->u.bfs_i.i_data.ds;

	while (pos >= 0) {
		struct buffer_head * bh;
		bfs_index_node *     bn;
		bfs_off_t            value;
		int                  flags;
		int                  k;
		
		if (pos)
			flags = 1;
		else
			flags = 0;

		iaddr = bfs_read_data_stream (sb, ds, &pos);
		if (!pos || !iaddr.len)
			break;

		/*
		 * read index node
		 */
		bh = bfs_read_index_node (iaddr, sb, flags);
		if (!bh)
			return -EBADF;

		bn = (bfs_index_node *) bh->b_data;
		BFS_DUMP_INDEX_NODE (bn);

		/*
		 * Is there no directry key in this index node?
		 */
		if (bn->all_key_count + count <= filp->f_pos) {
			/*
			 * Not exist
			 */
			count += bn->all_key_count;
			brelse(bh);
			continue;
		}

		k = filp->f_pos - count;

		/*
		 * search key
		 */
		if (!bfs_get_key_from_index_node (bn, &k, name, &len, &value)) {
			brelse (bh);
			return -EBADF;
		}

		error = filldir (dirent, name, len, filp->f_pos,
			(ino_t) value);

		if (error)
			filp->f_pos = 0;
		else
			filp->f_pos++;

		brelse (bh);
		return 0;
	}

	filp->f_pos = 0;
	return 0;
}
