/*
 * linux/fs/bfs/inode.c
 *
 * Copyright (C) 1999 Makoto Kato (m_kato@ga2.so-net.ne.jp)
 */

#include <asm/uaccess.h>
#include <asm/system.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/bfs_fs.h>
#include <linux/bfs_fs_sb.h>
#include <linux/bfs_fs_i.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/mm.h>



/*
 * return blocknumber
 */
int bfs_bmap (struct inode * inode, int block)
{
	struct bfs_sb_info    bs = inode->i_sb->u.bfs_sb;
	struct bfs_inode_info bi = inode->u.bfs_i;

	BFS_OUTPUT (("---> Enter bfs_bmap\n"));

	return  (bi.i_inode_num.allocation_group << bs.ag_shift)
		+ bi.i_inode_num.start;
}



/*
 * get buffer_header for inode.
 */

struct buffer_head * bfs_bread (dev_t dev, struct inode * inode)
{
	bfs_off_t            offset;
	struct super_block * sb = inode->i_sb;

	BFS_OUTPUT (("---> Enter bfs_read\n"));

	offset = (inode->u.bfs_i.i_inode_num.allocation_group
		<< sb->u.bfs_sb.ag_shift) + inode->u.bfs_i.i_inode_num.start;

	return bread (dev, offset, sb->s_blocksize);
}


struct buffer_head * bfs_bread2 (struct super_block * sb, bfs_inode_addr inode)
{
	dev_t     dev = sb->s_dev;
	bfs_off_t offset;

	BFS_OUTPUT (("---> Enter bfs_read2\n"));

	offset = (inode.allocation_group << sb->u.bfs_sb.ag_shift)
		+ inode.start;
	
	return bread (dev, offset, sb->s_blocksize);
}


void bfs_read_inode (struct inode * inode)
{
	struct buffer_head * bh;
	bfs_inode *          raw_inode;

	BFS_OUTPUT (("---> Enter bfs_read_inode  inode %ld[%lu,%u,%u]\n",
		inode->i_ino,
		inode->u.bfs_i.i_inode_num.allocation_group,
		inode->u.bfs_i.i_inode_num.start,
		inode->u.bfs_i.i_inode_num.len));

	/*
	 * convert from vfs's inode number to bfs's inode number
	 */
	if (inode->i_ino != BFSI2INODE(&(inode->u.bfs_i.i_inode_num),
		&(inode->i_sb->u.bfs_sb)))
	{
		inode->u.bfs_i.i_inode_num.allocation_group =
			inode->i_ino >> inode->i_sb->u.bfs_sb.ag_shift;
		inode->u.bfs_i.i_inode_num.start = inode->i_ino
				& ((2 << inode->i_sb->u.bfs_sb.ag_shift) - 1);
		inode->u.bfs_i.i_inode_num.len = 1; /* dummy */
	}

	if (!(bh = bfs_bread (inode->i_dev, inode))) {
		printk ("BFS: Unable to read inode block - inode=%ld",
			inode->i_ino);
		goto bad_inode;
	}

	raw_inode = (bfs_inode *) bh->b_data;

	BFS_DUMP_INODE (raw_inode);

	if (raw_inode->magic1 != BFS_INODE_MAGIC1) {
		printk ("BFS: inode = %ld is bad magic header.\n", inode->i_ino);
		brelse(bh);
		goto bad_inode;
	}

	inode->i_mode = (umode_t) raw_inode->mode;
	inode->i_uid = inode->i_sb->u.bfs_sb.mount_opts.uid ?
		inode->i_sb->u.bfs_sb.mount_opts.uid : (uid_t) raw_inode->uid;
	inode->i_gid = inode->i_sb->u.bfs_sb.mount_opts.gid ?
		inode->i_sb->u.bfs_sb.mount_opts.gid : (gid_t) raw_inode->gid;
	inode->i_nlink = 1;
	inode->i_size = raw_inode->data.datastream.size;

	/*
	 * BFS don't have access time.  So use last modified time.
	 */
	inode->i_atime = (time_t) (raw_inode->last_modified_time >> 16);

	/*
	 * BFS's time is 64 bit.
	 */
	inode->i_ctime = (time_t) (raw_inode->create_time >> 16);
	inode->i_mtime = (time_t) (raw_inode->last_modified_time >> 16);

	inode->i_blksize = raw_inode->inode_size;

	inode->i_blocks = raw_inode->inode_size / inode->i_sb->s_blocksize;
	inode->i_version = ++event;
	inode->u.bfs_i.i_inode_num = (bfs_inode_addr) raw_inode->inode_num;
	inode->u.bfs_i.i_mode = raw_inode->mode;
	inode->u.bfs_i.i_flags = raw_inode->flags;
	inode->u.bfs_i.i_create_time = raw_inode->create_time;
	inode->u.bfs_i.i_last_modified_time = raw_inode->last_modified_time;

	inode->u.bfs_i.i_parent = (bfs_inode_addr) raw_inode->parent;
	inode->u.bfs_i.i_attribute = (bfs_inode_addr) raw_inode->attributes;

	/*
	 * Symbolic link have no data stream.
	 * This position is original filename.
	 */
	if (!S_ISLNK(inode->i_mode))
		inode->u.bfs_i.i_data.ds = raw_inode->data.datastream;
	else {
		memcpy (inode->u.bfs_i.i_data.symlink, raw_inode->data.symlink,
			BFS_SYMLINK_LEN);
	}

	if (S_ISREG(inode->i_mode))
		inode->i_op = &bfs_file_inode_operations;
	else if (S_ISDIR(inode->i_mode))
		inode->i_op = &bfs_dir_inode_operations;
	else if (S_ISLNK(inode->i_mode))
		inode->i_op = &bfs_symlink_inode_operations;
	else if (S_ISCHR(inode->i_mode))
		inode->i_op = &chrdev_inode_operations;
	else if (S_ISBLK(inode->i_mode))
		inode->i_op = &blkdev_inode_operations;
	else if (S_ISFIFO(inode->i_mode))
		init_fifo(inode);

	brelse (bh);
	return;

bad_inode:
	make_bad_inode(inode);
	return;
}
