/*
 * linux/fs/bfs/namei.c
 *
 * Copyright 1999  Makoto Kato <m_kato@ga2.so-net.ne.jp>
 */

#include <asm/uaccess.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/bfs_fs.h>
#include <linux/fcntl.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/quotaops.h>


/*
 * find file or directry entry
 *
 *  dir     -> parent's i-node
 *  name    -> file or directry name
 *  namelen -> length of name
 */
static bfs_off_t bfs_find_entry (struct inode * dir, const char * const name,
	int namelen)
{
	struct buffer_head *    bh;
	bfs_index_node *        bn;
	struct super_block *    sb = dir->i_sb;
	int                     sd_pos = 0;
	int                     key_pos;
	char                    key[BFS_NAME_LEN + 1];
	int                     key_len;
	bfs_off_t               value;

	BFS_OUTPUT (("---> Enter bfs_find_entry\n"));

	while (sd_pos >= 0) {
		bfs_data_stream * ds = &dir->u.bfs_i.i_data.ds;
		bfs_inode_addr    iaddr;
		int               flags;

		/*
		 * get index node from small_data.
		 *
		 * First data stream has index entry, but other than it,
		 * doesn't have it.  
		 */
		if (sd_pos)
			flags = 1;
		else
			flags = 0;

		iaddr = bfs_read_data_stream (sb, ds, &sd_pos);
		if (!sd_pos)
			break;

		bh = bfs_read_index_node (iaddr, sb, flags);
		if (!bh) {
			BFS_OUTPUT (("  read_index_node faile\n"));
			return 0;
		}

		bn = (bfs_index_node *) bh->b_data;

		/*
		 * explore in this block
		 */
		key_pos = 0;
		while (key_pos >= 0) {
			bfs_get_key_from_index_node (bn, &key_pos, key,
				&key_len, &value);
			if (key_pos < 0) {
				BFS_OUTPUT (("  not exist in this block\n"));
				break;
			}

			if (strcmp (key, name) == 0 ) {
				brelse (bh);
				return value;
			}
		}

		brelse (bh);
	}
	return 0;
}



int bfs_lookup (struct inode * dir, struct dentry *dentry)
{
	struct inode * inode = NULL;
	bfs_off_t      offset;

	BFS_OUTPUT (("---> Enter bfs_lookup()  "));
	BFS_OUTPUT (("name %s inode %ld\n", dentry->d_name.name, dir->i_ino));

	if (dentry->d_name.len > BFS_NAME_LEN)
		return -ENAMETOOLONG;

	offset = bfs_find_entry (dir, dentry->d_name.name, dentry->d_name.len);
	if (offset) {
		inode = iget (dir->i_sb, (ino_t) offset); 
		if (!inode)
			return -EACCES;
	}

	d_add (dentry, inode);

	return 0;
}
